//--------------------------------------------------------------------------
//
//   Copyright (c) 2002, Colin Granville
//
//   All rights reserved.
//
//   Redistribution and use in source and binary forms, with or
//   without modification, are permitted provided that the following
//   conditions are met:
//
//      * Redistributions of source code must retain the above copyright
//        notice, this list of conditions and the following disclaimer.
//
//      * Redistributions in binary form must reproduce the above
//        copyright notice, this list of conditions and the following
//        disclaimer in the documentation and/or other materials
//        provided with the distribution.
//
//      * The name Colin Granville may not be used to endorse or promote
//        products derived from this software without specific prior
//        written permission.
//
//   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
//   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
//   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
//   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
//   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
//   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
//   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
//   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
//   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
//   OF THE POSSIBILITY OF SUCH DAMAGE.
//
//--------------------------------------------------------------------------

#include "GuiFlexFlex.h"
#include "kernel.h"
#include <fstream>
#include "flex.h"
#include "swis.h"

class GuiFlexFlex : public GuiFlex
{
  public:
    GuiFlexFlex(const char *program_name, int* error_fd);
    virtual         ~GuiFlexFlex();
    virtual bool    alloc(GuiFlexAnchor&,size_t n);
    virtual bool    free(GuiFlexAnchor&);
    virtual size_t  size(const GuiFlexAnchor&) const;
    virtual bool    extend(GuiFlexAnchor&,size_t new_size);
    virtual bool    midExtend(GuiFlexAnchor&,size_t at,int by);
    virtual bool    reanchor(GuiFlexAnchor& to, GuiFlexAnchor& from);
    virtual bool    setBudge(bool newstate);
    virtual void    saveHeapInfo(char* filename);
    virtual bool    compact();
    virtual bool    setDeferredCompaction(bool newstate);
    virtual size_t  availableMemory();
};

static GuiFlex* flex;


GuiFlex* GuiFlexFlex_init(const char *program_name, int *error_fd)
{
  if (!flex) flex=new GuiFlexFlex(program_name, error_fd);
  return flex;
}

//**************************************************************************

GuiFlexFlex::GuiFlexFlex(const char* program_name, int *error_fd)
{

#ifndef __APCS_32
  flex_init((char*)program_name,error_fd,134217728);
#else
  flex_init((char*)program_name,error_fd,0);
#endif
}

//**************************************************************************

GuiFlexFlex::~GuiFlexFlex()
{
}

//**************************************************************************

bool GuiFlexFlex::alloc(GuiFlexAnchor& anchor,size_t n)
{
  return flex_alloc((flex_ptr)&anchor.data, n);
}

//**************************************************************************

bool GuiFlexFlex::free(GuiFlexAnchor& anchor)
{
  if (anchor.data==0) return 1;
  flex_free((flex_ptr)&anchor.data);
  return anchor.data==0;
}

//**************************************************************************

size_t GuiFlexFlex::size(const GuiFlexAnchor& anchor) const
{
  if (anchor.data==0) return 0;
  return flex_size((flex_ptr)&anchor.data);
}

//**************************************************************************

bool GuiFlexFlex::extend(GuiFlexAnchor& anchor,size_t new_size)
{
  if (anchor.data==0) return 0;
  return flex_extend((flex_ptr)&anchor.data,new_size);
}

//**************************************************************************

bool GuiFlexFlex::midExtend(GuiFlexAnchor& anchor,size_t at,int by)
{
  if (anchor.data==0) return 0;
  return flex_midextend((flex_ptr)&anchor.data,at,by);
}

//**************************************************************************

bool GuiFlexFlex::reanchor(GuiFlexAnchor& to, GuiFlexAnchor& from)
{
#ifndef __APCS_32
  if (from.data==0) return 0;
  size_t* f = (size_t*)from.data;
  if (f[-2] != (size_t)&from.data) return 0;

  to.data=from.data;
  f[-2]=(size_t)&to.data;
  from.data=0;
  return 1;
#else
  return flex_reanchor((flex_ptr)&to.data,(flex_ptr)&from.data);
#endif
}

//**************************************************************************

bool GuiFlexFlex::setBudge(bool on)
{
#ifndef __APCS_32
  return (flex_budge == _kernel_register_slotextend(on?flex_budge:flex_dont_budge));
#else
  return flex_set_budge(on);
#endif

}

//**************************************************************************

void GuiFlexFlex::saveHeapInfo(char* filename)
{
#ifndef __APCS_32
  std::ofstream out(filename);
  if (!out) return;
  out << "GuiFlexFlex::saveHeapInfo not implemented";
#else
  flex_save_heap_info(filename);
#endif
}

//**************************************************************************

bool GuiFlexFlex::compact()
{
#ifndef __APCS_32
  return 0;
#else
  return flex_compact();
#endif
}

//**************************************************************************

bool GuiFlexFlex::setDeferredCompaction(bool on)
{
#ifndef __APCS_32
  return on;
#else
  return flex_set_deferred_compaction(on);
#endif
}

//**************************************************************************

size_t GuiFlexFlex::availableMemory()
{
  size_t slotsize=0;
  size_t maxsize=0x1BF8000;
  size_t freesize=0;
  _swix(Wimp_SlotSize,_INR(0,1) |_OUT(0) | _OUT(2),-1,-1,&slotsize, &freesize);
  _swix(OS_ReadDynamicArea,_IN(0) | _OUT(2),-1,&maxsize);
  if (freesize+slotsize>maxsize) freesize=maxsize-slotsize;
  return freesize;
}
