//--------------------------------------------------------------------------
//
//   Copyright (c) 2002, Colin Granville
//
//   All rights reserved.
//
//   Redistribution and use in source and binary forms, with or
//   without modification, are permitted provided that the following 
//   conditions are met:
//
//      * Redistributions of source code must retain the above copyright 
//        notice, this list of conditions and the following disclaimer.
//
//      * Redistributions in binary form must reproduce the above 
//        copyright notice, this list of conditions and the following 
//        disclaimer in the documentation and/or other materials 
//        provided with the distribution.
//
//      * The name Colin Granville may not be used to endorse or promote 
//        products derived from this software without specific prior 
//        written permission.
//
//   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
//   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
//   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
//   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
//   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
//   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
//   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
//   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
//   OF THE POSSIBILITY OF SUCH DAMAGE.
//
//--------------------------------------------------------------------------

#ifndef DocView_h
#define DocView_h

#include <memory>
#include "GuiWindow.h"
#include "GuiMenu.h"
#include "GuiGadget.h"
#include "GuiTargets.h"
#include "DocViewToolbar.h"
#include "Link.h"
#include "DrawOutputDevice.h"
#include "Save.h"
#include "DocViewChoices.h"
#include "DocPage.h"
#include "Node.h"

class HighlightBBox
{
  public:
    double xmin,ymin,xmax,ymax;
};

class Document;

class DragWorkArea
{
  public:
    void drag(GuiWindow& win);

  private:
    GuiWindow* window;
    int x;
    int y;

    GUI_DECLARE_EVENT_TARGETS(DragWorkArea);

    GuiWimpTarget dragEndedTarget;
    Claim dragEnded(GuiWimpPollBlock&,const GuiIdBlock&);

    GuiNullTarget nullEventTarget;
    void nullEvent();
    
};

//*************************************************************************

class DocView : public Node
{
  public:
    friend bool       makeDocView(Document*,const DocViewChoices&,int page=1,int logicalPageOffset=0);
    void              showPage();
    int               getPage()               {return toolbar.getPage();}
    void              setPage(const DocPage&,bool add_to_history=1);
    void              doAction(LinkAction* action,bool adjust);

    void              setHighlight(int pageNum,HighlightBBox&);
    void              clearHighlight();

    Document&         getDocument()           {return *document;}
    DrawOutputDevice& getOutDev()             {return outDev;}
    void              setTitle(const char*s)  {mainWin.setTitle(s);}
    enum              ScrollType {SCROLL_UP,SCROLL_DOWN,SCROLL_BOTTOM,SCROLL_TOP};
    bool              scroll(ScrollType);

    const DocViewChoices& getChoices()        {return choices;}
    GuiObject&            getWindow()         {return mainWin;}

    DECLARE_RTTI();

  private:
    DocView(Document&, const DocViewChoices&,int page,int logicalPageOffset);
    ~DocView(); //placed here because it must be created on heap - closeWindow deletes
    DocViewChoices   choices;
    Document*        document;
    GuiWindow        mainWin;
    GuiMenu          displayMenu;
    DocViewToolbar   toolbar;
    Save             save;
    DrawOutputDevice outDev;
    auto_ptr<Links>  links;
    bool             handPointerShowing;
    DragWorkArea     dragView;
    
    bool             highlightIsOn;
    HighlightBBox    highlightArea;
    void             transformHighlight(GuiBBox& out);

    void             set(Document& doc,const DocViewChoices&,int page,
                         int logicalPageOffset=0,bool add_to_history=1);

    void   setHandPointer(bool hand);
    void   Redraw(int);
    void   Scale(int);
    int    Scale();

    GUI_DECLARE_EVENT_TARGETS(DocView);

    GuiWimpTarget closeWindowTarget;
    Claim closeWindow(GuiWimpPollBlock&,const GuiIdBlock&);

    GuiWimpTarget openWindowTarget;
    Claim openWindow(GuiWimpPollBlock&,const GuiIdBlock&);

    GuiWimpTarget redrawWindowTarget;
    Claim redrawWindow(GuiWimpPollBlock&,const GuiIdBlock&);

    GuiWimpTarget mouseClickTarget;
    Claim mouseClick(GuiWimpPollBlock&,const GuiIdBlock&);

    GuiWimpTarget pointerEnteringTarget;
    Claim pointerEntering(GuiWimpPollBlock&,const GuiIdBlock&);

    GuiWimpTarget pointerLeavingTarget;
    Claim pointerLeaving(GuiWimpPollBlock&,const GuiIdBlock&);

    GuiNullTarget nullEventTarget;
    void nullEvent();

    GuiToolboxTarget newViewTarget;
    Claim newView(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget rotateViewTarget;
    Claim rotateView(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget printDialogOpenTarget;
    Claim printDialogOpen(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget displayMenuATBSTarget;
    Claim displayMenuATBS(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget miscMenuATBSTarget;
    Claim miscMenuATBS(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget saveMenuATBSTarget;
    Claim saveMenuATBS(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget highlightLinksTarget;
    Claim highlightLinks(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget setNoImagesTarget;
    Claim setNoImages(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget setNoTextTarget;
    Claim setNoText(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget setNoType3FontsTarget;
    Claim setNoType3Fonts(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget setNoDrawingsTarget;
    Claim setNoDrawings(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget fileInfoATBSTarget;
    Claim fileInfoATBS(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget scaleViewTarget;
    Claim scaleView(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget scaleUpTarget;
    Claim scaleUp(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget scaleDownTarget;
    Claim scaleDown(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget scale100Target;
    Claim scale100(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget scaleATBSTarget;
    Claim scaleATBS(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget saveDisplayChoicesTarget;
    Claim saveDisplayChoices(GuiToolboxEvent&,const GuiIdBlock&);

    GuiToolboxTarget findATBSTarget;
    Claim findATBS(GuiToolboxEvent&,const GuiIdBlock&);
};

#endif
