//--------------------------------------------------------------------------
//
//   Copyright (c) 2002, Colin Granville
//
//   All rights reserved.
//
//   Redistribution and use in source and binary forms, with or
//   without modification, are permitted provided that the following 
//   conditions are met:
//
//      * Redistributions of source code must retain the above copyright 
//        notice, this list of conditions and the following disclaimer.
//
//      * Redistributions in binary form must reproduce the above 
//        copyright notice, this list of conditions and the following 
//        disclaimer in the documentation and/or other materials 
//        provided with the distribution.
//
//      * The name Colin Granville may not be used to endorse or promote 
//        products derived from this software without specific prior 
//        written permission.
//
//   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
//   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
//   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
//   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
//   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
//   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
//   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
//   STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
//   OF THE POSSIBILITY OF SUCH DAMAGE.
//
//--------------------------------------------------------------------------

#include "PageHistory.h"
#include "DocView.h"
#include "Document.h"
#include "PageHistory.h"
#include "UserEvents.h"

PageHistory::PageHistory(DocView& v,DocViewToolbar& tb,GuiWindow& ancestor)
 : view(v),
   backIcon(tb,0x85),
   forwardIcon(tb,0x86),
   adjusterClickedTarget(&ancestor,GuiAdjuster::Clicked::Event,this,(Claim (PageHistory::*)(GuiToolboxEvent&, const GuiIdBlock&))&PageHistory::adjusterClicked),
   historyBackTarget(&ancestor,User_HistoryBack,this,(Claim (PageHistory::*)(GuiToolboxEvent&, const GuiIdBlock&))&PageHistory::historyBack),
   historyForwardTarget(&ancestor,User_HistoryForward,this,(Claim (PageHistory::*)(GuiToolboxEvent&, const GuiIdBlock&))&PageHistory::historyForward)
{
  begin=0;
  end=0;
  current=-1;
  pendingPage=0;
}

//****************************************************************

PageHistory::~PageHistory() {}

//****************************************************************

void PageHistory::add(int num)
{
  addPending();

  current++;
  if (current==PAGE_HISTORY_SIZE) current=0;

  buf[current].page=num;
  buf[current].docName=view.getDocument().getFileName();

  end=current+1;
  if (end==PAGE_HISTORY_SIZE) end=0;

  if (end==begin)
  {
    begin=end+1;
    if (begin==PAGE_HISTORY_SIZE) begin=0;
  }

  setFade();
}

//****************************************************************

const DocPage* PageHistory::back()
{
  addPending();

  if (current<0 || current==begin) return 0;
  current--;
  if (current<0) current+=PAGE_HISTORY_SIZE;
  setFade();
  return &buf[current];

}

//****************************************************************

const DocPage* PageHistory::forward()
{
  addPending();

  if (current<0) return 0;

  int next=current+1;
  if (next==PAGE_HISTORY_SIZE) next=0;
  if (next==end) return 0;
  current=next;
  setFade();
  return &buf[current];
}

//****************************************************************

void PageHistory::addPending(int num)
{
 pendingPage=num;
 forwardIcon.setFlags(GuiToolboxGadget_Faded);
 backIcon.setFlags(0);
}

//****************************************************************

void PageHistory::addPending()
{
  if (pendingPage==0) return;
  int page=pendingPage;
  pendingPage=0;
  add(page);
}

//*************************************************************************

void PageHistory::setFade()
{
  backIcon.setFlags((current<0 || current==begin) ? GuiToolboxGadget_Faded : 0);
  int next=current+1;
  if (next==PAGE_HISTORY_SIZE) next=0;
  forwardIcon.setFlags((current<0 || next==end) ? GuiToolboxGadget_Faded : 0);
}

//*************************************************************************

void PageHistory::setPage(const DocPage* page)
{
  if (!page) return;
  view.setPage(*page,0);
}

//*************************************************************************

Claim PageHistory::historyBack(GuiToolboxEvent&,const GuiIdBlock&)
{
  setPage(back());
  return CLAIM;
}

//*************************************************************************

Claim PageHistory::historyForward(GuiToolboxEvent&,const GuiIdBlock&)
{
  setPage(forward());
  return CLAIM;
}
//*************************************************************************

Claim PageHistory::adjusterClicked(GuiToolboxEvent& ev,const GuiIdBlock& idBlock)
{
  if (! (idBlock.self.component == backIcon.componentId() ||
         idBlock.self.component == forwardIcon.componentId()) )return DONT_CLAIM;

  GuiAdjuster::Clicked& click = (GuiAdjuster::Clicked&)ev;

  if (click.isDown()) 
     historyForward(ev,idBlock);
  else
     historyBack(ev,idBlock);
     
  return CLAIM;
};

